'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (CoreManager, Exports) {

  Exports.run =

  /**
   * Contains functions for calling and declaring
   * <a href="/docs/cloud_code_guide#functions">cloud functions</a>.
   * <p><strong><em>
   *   Some functions are only available from Cloud Code.
   * </em></strong></p>
   *
   * @class Parse.Cloud
   * @static
   * @hideconstructor
   */

  /**
   * Makes a call to a cloud function.
   * @method run
   * @name Parse.Cloud.run
   * @param {String} name The function name.
   * @param {Object} data The parameters to send to the cloud function.
   * @param {Object} options A Backbone-style options object
   * options.success, if set, should be a function to handle a successful
   * call to a cloud function.  options.error should be a function that
   * handles an error running the cloud function.  Both functions are
   * optional.  Both functions take a single argument.
   * @return {Parse.Promise} A promise that will be resolved with the result
   * of the function.
   */
  function (name, data, options) {
    options = options || {};

    if (typeof name !== 'string' || name.length === 0) {
      throw new TypeError('Cloud function name must be a string.');
    }

    var requestOptions = {};
    if (options.useMasterKey) {
      requestOptions.useMasterKey = options.useMasterKey;
    }
    if (options.sessionToken) {
      requestOptions.sessionToken = options.sessionToken;
    }

    return CoreManager.getCloudController().run(name, data, requestOptions)._thenRunCallbacks(options);
  };

  var decode = (0, _include2.default)('./decode', CoreManager);
  var encode = (0, _include2.default)('./encode', CoreManager);
  var ParseError = (0, _include2.default)('./ParseError', CoreManager);
  var ParsePromise = (0, _include2.default)('./ParsePromise', CoreManager);

  CoreManager.setCloudController({
    run: function (name, data, options) {
      var RESTController = CoreManager.getRESTController();

      var payload = encode.default(data, true);

      var requestOptions = {};
      if (options.hasOwnProperty('useMasterKey')) {
        requestOptions.useMasterKey = options.useMasterKey;
      }
      if (options.hasOwnProperty('sessionToken')) {
        requestOptions.sessionToken = options.sessionToken;
      }

      var request = RESTController.request('POST', 'functions/' + name, payload, requestOptions);

      return request.then(function (res) {
        var decoded = decode.default(res);
        if (decoded && decoded.hasOwnProperty('result')) {
          return ParsePromise.default.as(decoded.result);
        }
        return ParsePromise.default.error(new ParseError.default(ParseError.default.INVALID_JSON, 'The server returned an invalid response.'));
      })._thenRunCallbacks(options);
    }
  });
};

var _include = require('./include');

var _include2 = _interopRequireDefault(_include);

function _interopRequireDefault(obj) {
  return obj && obj.__esModule ? obj : { default: obj };
}