/*
 * Copyright (c) 2015-present, Parse, LLC.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 * 
 */

function requireMethods(name, methods, controller) {
  methods.forEach(func => {
    if (typeof controller[func] !== 'function') {
      throw new Error(`${name} must implement ${func}()`);
    }
  });
}

export default class CoreManagerType {

  constructor() {
    this.config = {
      // Defaults
      IS_NODE: typeof process !== 'undefined' && !!process.versions && !!process.versions.node && !process.versions.electron,
      REQUEST_ATTEMPT_LIMIT: 5,
      SERVER_URL: 'https://api.parse.com/1',
      LIVEQUERY_SERVER_URL: null,
      VERSION: 'js' + '2.0.3',
      APPLICATION_ID: null,
      JAVASCRIPT_KEY: null,
      MASTER_KEY: null,
      USE_MASTER_KEY: false,
      PERFORM_USER_REWRITE: true,
      FORCE_REVOCABLE_SESSION: false
    };
  }

  get(key) {
    if (this.config.hasOwnProperty(key)) {
      return this.config[key];
    }
    throw new Error('Configuration key not found: ' + key);
  }

  set(key, value) {
    this.config[key] = value;
  }

  /* Specialized Controller Setters/Getters */

  setAnalyticsController(controller) {
    requireMethods('AnalyticsController', ['track'], controller);
    this.config['AnalyticsController'] = controller;
  }

  getAnalyticsController() {
    return this.config['AnalyticsController'];
  }

  setCloudController(controller) {
    requireMethods('CloudController', ['run'], controller);
    this.config['CloudController'] = controller;
  }

  getCloudController() {
    return this.config['CloudController'];
  }

  setConfigController(controller) {
    requireMethods('ConfigController', ['current', 'get'], controller);
    this.config['ConfigController'] = controller;
  }

  getConfigController() {
    return this.config['ConfigController'];
  }

  setFileController(controller) {
    requireMethods('FileController', ['saveFile', 'saveBase64'], controller);
    this.config['FileController'] = controller;
  }

  getFileController() {
    return this.config['FileController'];
  }

  setInstallationController(controller) {
    requireMethods('InstallationController', ['currentInstallationId'], controller);
    this.config['InstallationController'] = controller;
  }

  getInstallationController() {
    return this.config['InstallationController'];
  }

  setObjectController(controller) {
    requireMethods('ObjectController', ['save', 'fetch', 'destroy'], controller);
    this.config['ObjectController'] = controller;
  }

  getObjectController() {
    return this.config['ObjectController'];
  }

  setObjectStateController(controller) {
    requireMethods('ObjectStateController', ['getState', 'initializeState', 'removeState', 'getServerData', 'setServerData', 'getPendingOps', 'setPendingOp', 'pushPendingState', 'popPendingState', 'mergeFirstPendingState', 'getObjectCache', 'estimateAttribute', 'estimateAttributes', 'commitServerChanges', 'enqueueTask', 'clearAllState'], controller);

    this.config['ObjectStateController'] = controller;
  }

  getObjectStateController() {
    return this.config['ObjectStateController'];
  }

  setPushController(controller) {
    requireMethods('PushController', ['send'], controller);
    this.config['PushController'] = controller;
  }

  getPushController() {
    return this.config['PushController'];
  }

  setQueryController(controller) {
    requireMethods('QueryController', ['find', 'aggregate'], controller);
    this.config['QueryController'] = controller;
  }

  getQueryController() {
    return this.config['QueryController'];
  }

  setRESTController(controller) {
    requireMethods('RESTController', ['request', 'ajax'], controller);
    this.config['RESTController'] = controller;
  }

  getRESTController() {
    return this.config['RESTController'];
  }

  setSchemaController(controller) {
    requireMethods('SchemaController', ['get', 'create', 'update', 'delete', 'send', 'purge'], controller);
    this.config['SchemaController'] = controller;
  }

  getSchemaController() {
    return this.config['SchemaController'];
  }

  setSessionController(controller) {
    requireMethods('SessionController', ['getSession'], controller);
    this.config['SessionController'] = controller;
  }

  getSessionController() {
    return this.config['SessionController'];
  }

  setStorageController(controller) {
    if (controller.async) {
      requireMethods('An async StorageController', ['getItemAsync', 'setItemAsync', 'removeItemAsync'], controller);
    } else {
      requireMethods('A synchronous StorageController', ['getItem', 'setItem', 'removeItem'], controller);
    }
    this.config['StorageController'] = controller;
  }

  getStorageController() {
    return this.config['StorageController'];
  }

  setAsyncStorage(storage) {
    this.config['AsyncStorage'] = storage;
  }

  getAsyncStorage() {
    return this.config['AsyncStorage'];
  }

  setUserController(controller) {
    requireMethods('UserController', ['setCurrentUser', 'currentUser', 'currentUserAsync', 'signUp', 'logIn', 'become', 'logOut', 'requestPasswordReset', 'upgradeToRevocableSession', 'linkWith'], controller);
    this.config['UserController'] = controller;
  }

  getUserController() {
    return this.config['UserController'];
  }

  setLiveQueryController(controller) {
    requireMethods('LiveQueryController', ['subscribe', 'unsubscribe', 'open', 'close'], controller);
    this.config['LiveQueryController'] = controller;
  }

  getLiveQueryController() {
    return this.config['LiveQueryController'];
  }

  setHooksController(controller) {
    requireMethods('HooksController', ['create', 'get', 'update', 'remove'], controller);
    this.config['HooksController'] = controller;
  }

  getHooksController() {
    return this.config['HooksController'];
  }

  setTasksController(controller) {
    requireMethods('TasksController', ['create', 'get', 'update', 'remove'], controller);
    this.config['TasksController'] = controller;
  }

  getTasksController() {
    return this.config['TasksController'];
  }
}

export const global = new CoreManagerType();