/**
 * Copyright (c) 2015-present, Parse, LLC.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */
import include from './include';
import CoreManagerType from './CoreManager';

export default function createParse() {

  const CoreManager = new CoreManagerType();

  const decode = include('./decode', CoreManager);
  const encode = include('./encode', CoreManager);
  const InstallationController = include('./InstallationController', CoreManager);
  const ParseOp = include('./ParseOp', CoreManager);
  const RESTController = include('./RESTController', CoreManager);

  /**
   * Contains all Parse API classes and functions.
   * @static
   * @global
   * @class
   * @hideconstructor
   */
  var Parse = {
    /**
     * Call this method first to set up your authentication tokens for Parse.
     * You can get your keys from the Data Browser on parse.com.
     * @param {String} applicationId Your Parse Application ID.
     * @param {String} javaScriptKey (optional) Your Parse JavaScript Key (Not needed for parse-server)
     * @param {String} masterKey (optional) Your Parse Master Key. (Node.js only!)
     * @static
     */
    initialize(applicationId, javaScriptKey) {
      Parse._initialize(applicationId, javaScriptKey);
    },

    _initialize(applicationId, javaScriptKey, masterKey) {
      CoreManager.set('APPLICATION_ID', applicationId);
      CoreManager.set('JAVASCRIPT_KEY', javaScriptKey);
      CoreManager.set('MASTER_KEY', masterKey);
      CoreManager.set('USE_MASTER_KEY', false);
    },

    /**
     * Call this method to set your AsyncStorage engine
     * Starting Parse@1.11, the ParseSDK do not provide a React AsyncStorage as the ReactNative module
     * is not provided at a stable path and changes over versions.
     * @param {AsyncStorage} storage a react native async storage.
     * @static
     */
    setAsyncStorage(storage) {
      CoreManager.setAsyncStorage(storage);
    }
  };

  /** These legacy setters may eventually be deprecated **/
  /** 
   * @member Parse.applicationId
   * @type string
   * @static
   */
  Object.defineProperty(Parse, 'applicationId', {
    get() {
      return CoreManager.get('APPLICATION_ID');
    },
    set(value) {
      CoreManager.set('APPLICATION_ID', value);
    }
  });

  /** 
   * @member Parse.javaScriptKey
   * @type string
   * @static
   */
  Object.defineProperty(Parse, 'javaScriptKey', {
    get() {
      return CoreManager.get('JAVASCRIPT_KEY');
    },
    set(value) {
      CoreManager.set('JAVASCRIPT_KEY', value);
    }
  });

  /** 
   * @member Parse.masterKey
   * @type string
   * @static
   */
  Object.defineProperty(Parse, 'masterKey', {
    get() {
      return CoreManager.get('MASTER_KEY');
    },
    set(value) {
      CoreManager.set('MASTER_KEY', value);
    }
  });

  /** 
   * @member Parse.serverURL
   * @type string
   * @static
   */
  Object.defineProperty(Parse, 'serverURL', {
    get() {
      return CoreManager.get('SERVER_URL');
    },
    set(value) {
      CoreManager.set('SERVER_URL', value);
    }
  });
  /** 
   * @member Parse.liveQueryServerURL
   * @type string
   * @static
   */
  Object.defineProperty(Parse, 'liveQueryServerURL', {
    get() {
      return CoreManager.get('LIVEQUERY_SERVER_URL');
    },
    set(value) {
      CoreManager.set('LIVEQUERY_SERVER_URL', value);
    }
  });

  /**
   * @member Parse.datonisURL
   * @type string
   * @static
   */
  Object.defineProperty(Parse, 'datonisURL', {
    get() {
      return CoreManager.get('DATONIS_URL');
    },
    set(value) {
      CoreManager.set('DATONIS_URL', value);
    }
  });
  /* End setters */

  Parse.ACL = include('./ParseACL', CoreManager).default;
  Parse.Analytics = include('./Analytics', CoreManager);
  Parse.Cloud = include('./Cloud', CoreManager);
  Parse.CoreManager = CoreManager;
  Parse.Config = include('./ParseConfig', CoreManager).default;
  Parse.Error = include('./ParseError', CoreManager).default;
  Parse.FacebookUtils = include('./FacebookUtils', CoreManager).default;
  Parse.File = include('./ParseFile', CoreManager).default;
  Parse.GeoPoint = include('./ParseGeoPoint', CoreManager).default;
  Parse.Polygon = include('./ParsePolygon', CoreManager).default;
  Parse.Installation = include('./ParseInstallation', CoreManager).default;
  Parse.Object = include('./ParseObject', CoreManager).default;
  Parse.Op = {
    Set: ParseOp.SetOp,
    Unset: ParseOp.UnsetOp,
    Increment: ParseOp.IncrementOp,
    Add: ParseOp.AddOp,
    Remove: ParseOp.RemoveOp,
    AddUnique: ParseOp.AddUniqueOp,
    Relation: ParseOp.RelationOp
  };
  Parse.Promise = include('./ParsePromise', CoreManager).default;
  Parse.Push = include('./Push', CoreManager);
  Parse.Query = include('./ParseQuery', CoreManager).default;
  Parse.Relation = include('./ParseRelation', CoreManager).default;
  Parse.Role = include('./ParseRole', CoreManager).default;
  Parse.Schema = include('./ParseSchema', CoreManager).default;
  Parse.Session = include('./ParseSession', CoreManager).default;
  Parse.Storage = include('./Storage', CoreManager);
  Parse.User = include('./ParseUser', CoreManager).default;
  Parse.LiveQuery = include('./ParseLiveQuery', CoreManager).default;
  Parse.LiveQueryClient = include('./LiveQueryClient', CoreManager).default;
  //Setting Datonis object
  Parse.Datonis = include('./Datonis', CoreManager).default;
  Parse._request = function (...args) {
    return CoreManager.getRESTController().request.apply(null, args);
  };
  Parse._ajax = function (...args) {
    return CoreManager.getRESTController().ajax.apply(null, args);
  };
  // We attempt to match the signatures of the legacy versions of these methods
  Parse._decode = function (_, value) {
    return decode.default(value);
  };
  Parse._encode = function (value, _, disallowObjects) {
    return encode.default(value, disallowObjects);
  };
  Parse._getInstallationId = function () {
    return CoreManager.getInstallationController().currentInstallationId();
  };

  CoreManager.setInstallationController(InstallationController);
  CoreManager.setRESTController(RESTController);

  Parse.Hooks = include('./ParseHooks', CoreManager);
  Parse.Tasks = include('./ParseTasks', CoreManager);

  // For legacy requires, of the form `var Parse = include('@datonis/parse').Parse`
  Parse.Parse = Parse;

  return Parse;
}