/**
 * Copyright (c) 2015-present, Parse, LLC.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 * 
 */
import include from './include';
export default function (CoreManager, Exports) {

  Exports.getState = getState;
  Exports.initializeState = initializeState;
  Exports.removeState = function (obj) {
    let state = getState(obj);
    if (state === null) {
      return null;
    }
    objectState.delete(obj);
    return state;
  };
  Exports.getServerData = getServerData;
  Exports.setServerData = function (obj, attributes) {
    let serverData = initializeState(obj).serverData;
    ObjectStateMutations.setServerData(serverData, attributes);
  };
  Exports.getPendingOps = getPendingOps;
  Exports.setPendingOp = function (obj, attr, op) {
    let pendingOps = initializeState(obj).pendingOps;
    ObjectStateMutations.setPendingOp(pendingOps, attr, op);
  };
  Exports.pushPendingState = function (obj) {
    let pendingOps = initializeState(obj).pendingOps;
    ObjectStateMutations.pushPendingState(pendingOps);
  };
  Exports.popPendingState = function (obj) {
    let pendingOps = initializeState(obj).pendingOps;
    return ObjectStateMutations.popPendingState(pendingOps);
  };
  Exports.mergeFirstPendingState = function (obj) {
    let pendingOps = getPendingOps(obj);
    ObjectStateMutations.mergeFirstPendingState(pendingOps);
  };
  Exports.getObjectCache = function (obj) {
    let state = getState(obj);
    if (state) {
      return state.objectCache;
    }
    return {};
  };
  Exports.estimateAttribute = function (obj, attr) {
    let serverData = getServerData(obj);
    let pendingOps = getPendingOps(obj);
    return ObjectStateMutations.estimateAttribute(serverData, pendingOps, obj.className, obj.id, attr);
  };
  Exports.estimateAttributes = function (obj) {
    let serverData = getServerData(obj);
    let pendingOps = getPendingOps(obj);
    return ObjectStateMutations.estimateAttributes(serverData, pendingOps, obj.className, obj.id);
  };
  Exports.commitServerChanges = function (obj, changes) {
    let state = initializeState(obj);
    ObjectStateMutations.commitServerChanges(state.serverData, state.objectCache, changes);
  };
  Exports.enqueueTask = function (obj, task) {
    let state = initializeState(obj);
    return state.tasks.enqueue(task);
  };
  Exports.duplicateState = function (source, dest) {
    let oldState = initializeState(source);
    let newState = initializeState(dest);
    for (let key in oldState.serverData) {
      newState.serverData[key] = oldState.serverData[key];
    }
    for (let index = 0; index < oldState.pendingOps.length; index++) {
      for (let key in oldState.pendingOps[index]) {
        newState.pendingOps[index][key] = oldState.pendingOps[index][key];
      }
    }
    for (let key in oldState.objectCache) {
      newState.objectCache[key] = oldState.objectCache[key];
    }
    newState.existed = oldState.existed;
  };
  Exports.clearAllState = function () {
    objectState = new WeakMap();
  };

  const ObjectStateMutations = include('./ObjectStateMutations', CoreManager);
  const TaskQueue = include('./TaskQueue', CoreManager);

  let objectState = new WeakMap();

  function getState(obj) {
    let classData = objectState.get(obj);
    return classData || null;
  }

  function initializeState(obj, initial) {
    let state = getState(obj);
    if (state) {
      return state;
    }
    if (!initial) {
      initial = {
        serverData: {},
        pendingOps: [{}],
        objectCache: {},
        tasks: new TaskQueue.default(),
        existed: false
      };
    }
    state = initial;
    objectState.set(obj, state);
    return state;
  }

  function getServerData(obj) {
    let state = getState(obj);
    if (state) {
      return state.serverData;
    }
    return {};
  }

  function getPendingOps(obj) {
    let state = getState(obj);
    if (state) {
      return state.pendingOps;
    }
    return [{}];
  }
}